import datetime
import typing

from google.api_core.datetime_helpers import DatetimeWithNanoseconds
from google.cloud.spanner_v1 import param_types, Type, JsonObject
from google.cloud.spanner_v1.streamed import StreamedResultSet


def zip_results(result_set: StreamedResultSet, field_names: typing.List[str] = None):
    """
    :param result_set: the results
    :param field_names: list of field names. If none is provided, then we use the column names.
    :return: a list of dict(s) zipped with {field_name/column_name: column_value}
    """
    _results = []
    for row in result_set:
        if field_names is None:
            field_names = [field.name for field in result_set.fields]
        _results.append(dict(zip(field_names, row)))
    return _results


def zip_first_result(
    result_set: StreamedResultSet, field_names: typing.List[str] = None
) -> typing.Optional[dict]:
    """
    :param result_set: the results
    :param field_names: list of field names. If none is provided, then we use the column names.
    :return: a single dict zipped with {field_name/column_name: column_value} or None
    """
    for row in result_set:
        if field_names is None:
            field_names = [field.name for field in result_set.fields]
        return dict(zip(field_names, row))
    return None


def single_result(result_set: StreamedResultSet) -> typing.Optional[typing.Any]:
    """
    :param result_set: the results
    :return: the first column value of the first row or None
    """
    for row in result_set:
        return row[0]
    return None


def _sanitize_value(v: typing.Any):
    if isinstance(v, dict):
        return JsonObject(v)
    return v


def _resolve_type(v: typing.Any) -> Type:
    if isinstance(v, int):
        return param_types.INT64
    if isinstance(v, float):
        return param_types.FLOAT64
    if isinstance(v, bool):
        return param_types.BOOL
    if isinstance(v, str):
        return param_types.STRING
    if isinstance(v, (datetime.datetime, DatetimeWithNanoseconds)):
        return param_types.TIMESTAMP
    if isinstance(v, datetime.date):
        return param_types.DATE
    if isinstance(v, dict):
        return param_types.JSON
    if isinstance(v, (list, set, tuple)):
        if not len(v):
            raise ValueError(f"Cannot resolve type of empty list/set/tuple")
        item_type = _resolve_type(v[0])
        return param_types.Array(item_type)
    raise ValueError(f"Unsupported type {type(v)}")


def build_params_ptypes(
    values: dict,
) -> (typing.Dict[str, typing.Any], typing.Dict[str, Type]):
    params = {}
    ptypes = {}
    for k, v in values.items():
        if v is not None:
            params.update({k: _sanitize_value(v)})
            ptypes.update({k: _resolve_type(v)})
    return (
        params,
        ptypes,
    )


def build_params_ptypes_as_kwargs(
    values: dict,
) -> (typing.Dict[str, typing.Any], typing.Dict[str, Type]):
    """
    Utility method to build `params` and `param_types` out of one single dict.
    Example:

        result_set = await snapshot.execute_sql(
            sql="SELECT * FROM Users WHERE user_id = @user_id",
            **build_params_ptypes_as_kwargs({"user_id": user_id}),
        )
    """
    params, ptypes = build_params_ptypes(values)
    return {
        "params": params,
        "param_types": ptypes,
    }



import _queue
import logging
from typing import TypeVar, Callable

from google.api_core import exceptions
from google.cloud import spanner
from google.cloud.spanner_v1.snapshot import Snapshot
from google.cloud.spanner_v1.transaction import Transaction

from sparta.spanner.concurrency import run_as_non_blocking

_T = TypeVar("_T")


class DBService:
    def __init__(
        self,
        instance_id: str,
        database_id: str,
        pool_size: int = None,
        session_request_timeout: int = None,
        spanner_client: spanner.Client = None,
        # unused, just for backward compatibility
        project_id: str = None,
    ):
        """
        :param instance_id: Spanner instance ID
        :param database_id: Spanner database ID
        :param pool_size: (optional) size param for spanner.FixedSizePool
        :param session_request_timeout: (optional) default_timeout param for spanner.FixedSizePool
        :param spanner_client: (optional) a spanner.Client instance
        """
        if not instance_id:
            raise ValueError("Missing instance_id")
        if not database_id:
            raise ValueError("Missing database_id")

        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")

        _pool = None
        if pool_size:
            _pool = spanner.FixedSizePool(
                size=pool_size, default_timeout=session_request_timeout
            )

        self.spanner_client = spanner_client or spanner.Client()
        self.instance = self.spanner_client.instance(instance_id)
        self.database = self.instance.database(database_id, pool=_pool)

        self.logger.debug(f"Created {type(self).__name__} for {self.database.name}")

    def ping(self):
        self.execute_sql("SELECT 1")

    def execute_sql(self, *arg, **kwarg):
        with self.database.snapshot() as s:
            return s.execute_sql(*arg, **kwarg)

    async def run_in_snapshot(
        self,
        snapshot_consumer: Callable[[Snapshot], _T],
        *args,
        **kwargs,
    ) -> _T:
        def task():
            try:
                with self.database.snapshot(*args, **kwargs) as snapshot:
                    return snapshot_consumer(snapshot)
            except _queue.Empty as error:
                raise NoSessionAvailableException() from error  # exception chaining

        return await run_as_non_blocking(task)

    async def run_in_transaction(
        self,
        transaction_consumer: Callable[[Transaction], _T],
        *args,
        **kwargs,
    ) -> _T:
        def task():
            try:
                return self.database.run_in_transaction(
                    transaction_consumer, *args, **kwargs
                )
            except _queue.Empty as error:
                raise NoSessionAvailableException() from error  # exception chaining
            except exceptions.Aborted as error:
                raise TimeoutError() from error  # exception chaining

        return await run_as_non_blocking(task)


class NoSessionAvailableException(Exception):
    pass



import dataclasses
import typing

from google.cloud.spanner_v1 import param_types

from sparta.spanner import DBService
from sparta.spanner.utils import zip_results

_LAST_UPDATED_AT = "last_updated_at"
_LAST_UPDATED_BY = "last_updated_by"
_CREATED_AT = "created_at"
_CREATED_BY = "created_by"
_UNMODIFIABLE_FIELDS = [_CREATED_AT, _CREATED_BY, _LAST_UPDATED_AT, _LAST_UPDATED_BY]


@dataclasses.dataclass
class ColumnMetadata:
    ordinal_position: int
    column_name: str
    spanner_type: str
    is_nullable: bool
    constraint_types: typing.List[str]


@dataclasses.dataclass
class TableMetadata:
    table_name: str
    table_pk: str
    table_pks: typing.List[str]
    columns_meta: typing.List[ColumnMetadata]
    unmodifiable_fields: typing.List[str]
    column_names: typing.List[str]


_SQL_METADATA = """SELECT ordinal_position, column_name, spanner_type, is_nullable, constraint_types
FROM (
    SELECT ordinal_position, column_name, spanner_type, is_nullable FROM information_schema.columns 
    WHERE table_name = @table_name
) LEFT OUTER JOIN (
    SELECT column_name, ARRAY_AGG(constraint_type) as constraint_types 
    FROM information_schema.KEY_COLUMN_USAGE
    JOIN information_schema.TABLE_CONSTRAINTS 
    USING (constraint_name, table_name)
    WHERE table_name = @table_name
    GROUP BY column_name
) USING (column_name)
ORDER BY ordinal_position"""


def _load_columns_meta(db: DBService, table_name: str) -> typing.List[ColumnMetadata]:
    assert table_name
    assert db
    params = {"table_name": table_name}
    ptypes = {
        "table_name": param_types.STRING,
    }
    result_set = db.execute_sql(
        sql=_SQL_METADATA,
        params=params,
        param_types=ptypes,
    )
    _list = zip_results(result_set)
    return [ColumnMetadata(**c) for c in _list]


def get_table_metadata(db: DBService, table_name: str) -> TableMetadata:
    assert table_name
    assert db
    columns_meta = _load_columns_meta(db, table_name)
    if not columns_meta:
        raise RuntimeError(f"No metadata for collection '{table_name}'")
    table_pks = [
        c.column_name
        for c in columns_meta
        if c.constraint_types and "PRIMARY KEY" in c.constraint_types
    ]
    if not table_pks:
        raise RuntimeError(f"Invalid metadata for collection '{table_name}'")
    elif len(table_pks) > 1:
        raise RuntimeError(
            f"Composite primary keys are not supported, yet (table_name: {table_name}, pks: {table_pks})"
        )
    unmodifiable_fields = table_pks + _UNMODIFIABLE_FIELDS
    column_names = [x.column_name for x in columns_meta]
    return TableMetadata(
        table_name=table_name,
        table_pk=table_pks[0],
        table_pks=table_pks,
        columns_meta=columns_meta,
        column_names=column_names,
        unmodifiable_fields=unmodifiable_fields,
    )



from google.cloud.spanner_v1.snapshot import Snapshot
from google.cloud.spanner_v1.streamed import StreamedResultSet
from google.cloud.spanner_v1.transaction import Transaction

from sparta.spanner import DBService


class DBServiceAsyncWrapper:
    def __init__(
        self,
        db: DBService,
    ):
        self.db = db

    async def execute_sql_async(self, *arg, **kwarg) -> StreamedResultSet:
        def func(snapshot: Snapshot):
            return snapshot.execute_sql(*arg, **kwarg)

        return await self.db.run_in_snapshot(func)

    async def read_async(self, *arg, **kwarg) -> StreamedResultSet:
        def func(snapshot: Snapshot):
            return snapshot.read(*arg, **kwarg)

        return await self.db.run_in_snapshot(func)

    async def update_async(self, *arg, **kwarg) -> StreamedResultSet:
        def func(transition: Transaction):
            return transition.update(*arg, **kwarg)

        return await self.db.run_in_transaction(func)

    async def insert_async(self, *arg, **kwarg) -> StreamedResultSet:
        def func(transition: Transaction):
            return transition.insert(*arg, **kwarg)

        return await self.db.run_in_transaction(func)

    async def insert_or_update_async(self, *arg, **kwarg) -> StreamedResultSet:
        def func(transition: Transaction):
            return transition.insert_or_update(*arg, **kwarg)

        return await self.db.run_in_transaction(func)

    async def delete_async(self, *arg, **kwarg) -> StreamedResultSet:
        def func(transition: Transaction):
            return transition.delete(*arg, **kwarg)

        return await self.db.run_in_transaction(func)



import dataclasses
import logging
import threading
import typing

from google.cloud import spanner

from sparta.spanner import DBService


@dataclasses.dataclass
class DBServiceConfig:
    instance_id: str
    database_id: str
    pool_size: typing.Optional[int] = None
    session_request_timeout: typing.Optional[int] = None


class DBServiceProvider:
    """
    Provides DBService instances.
    Responsible for creation and re-utilization of existing instances.
    """

    def __init__(self) -> None:
        super().__init__()
        self.logger = logging.getLogger(f"{__name__}.{type(self).__name__}")
        self._spanner_client = spanner.Client()
        self._instances = {}
        # Use lock to assure thread-safety.
        # See https://medium.com/analytics-vidhya/how-to-create-a-thread-safe-singleton-class-in-python-822e1170a7f6
        self._lock = threading.Lock()

    def get_by_config(self, config: DBServiceConfig) -> DBService:
        _key = hash(str(config))
        if _key not in self._instances:
            with self._lock:
                # Another thread could have created the instance before we acquired the lock. Double-check.
                if _key not in self._instances:
                    self._instances[_key] = DBService(
                        instance_id=config.instance_id,
                        database_id=config.database_id,
                        pool_size=config.pool_size,
                        session_request_timeout=config.session_request_timeout,
                        spanner_client=self._spanner_client,
                    )
        return self._instances[_key]

    def get(
        self,
        instance_id: str,
        database_id: str,
        pool_size: typing.Optional[int] = None,
        session_request_timeout: typing.Optional[int] = None,
    ) -> DBService:
        return self.get_by_config(
            DBServiceConfig(
                instance_id=instance_id,
                database_id=database_id,
                pool_size=pool_size,
                session_request_timeout=session_request_timeout,
            )
        )



import functools
import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, TypeVar

_T = TypeVar("_T")


async def run_as_non_blocking(func: Callable[[...], _T], *args, **kwargs) -> _T:
    try:
        from starlette.concurrency import run_in_threadpool

        return await run_in_threadpool(func, *args, **kwargs)
    except ModuleNotFoundError as error:
        with ThreadPoolExecutor(max_workers=1) as executor:
            return await asyncio.get_event_loop().run_in_executor(
                executor, functools.partial(func, *args, **kwargs)
            )
